{$APPTYPE GUI}
{$ifdef FPC} {$mode delphi} {$endif}

uses SDL, dglOpenGL, Shader, TGA;

const
  WINDOW_X = 1024;
  WINDOW_Y = 768;

type
  TSelObjekt = record
    displayliste: GLuInt;
    posx, posy: single;
    name: string;
  end;

var
  SelShader: record
    prog: glHandle;
    color: GLint;
    use_alpha_tex: GLint;
  end;

  Objekte: array [1..3] of TSelObjekt;


//-----------------------------------------------------------------
// Allgemeine Renderprozedur fr Selektion und normale Darstellung
//-----------------------------------------------------------------
procedure Render;
const
  fac: single = 1/255;
var
  i: longword;
  iv: array [0..3] of byte absolute i; // iv und i berschneiden sich
begin
  glClear(GL_COLOR_BUFFER_BIT);

  // Das Erste Model soll texturiert werden
  glEnable(GL_TEXTURE_2D);
  glUniform1f(SelShader.use_alpha_tex, 1.0);

  for i := 1 to high(Objekte) do
  begin
    if i = 2 then begin // Ab dem zweiten nicht mehr
      glDisable(GL_TEXTURE_2D);
      glUniform1f(SelShader.use_alpha_tex, 0.0);
    end;

    // Den Index in eine Farbe umrechnen und im Shader setzen
    {$ifdef ENDIAN_BIG}
    glUniform4f(SelShader.color, iv[1]*fac, iv[2]*fac, iv[3]*fac, 1.0);
    {$else}
    glUniform4f(SelShader.color, iv[0]*fac, iv[1]*fac, iv[2]*fac, 1.0);
    {$endif}

    // Zeichnen
    glLoadIdentity;
    glTranslatef(Objekte[i].posx, Objekte[i].posy, 0);
    glCallList(Objekte[i].displayliste);
  end;
end;


//-----------------------------------------------------------------
// Gibt den Index des angeklickten Objektes oder 0 zurck
//-----------------------------------------------------------------
function Select(x, y: integer): longword;
var
  PixelData: array [0..9] of longword;
  count, i: longword;
begin
  // Koordinaten umrechnen (SDL-Fenster in OpenGL / 3x3-Box zentrieren)
  y := WINDOW_Y-y;
  if x > 0 then if x < WINDOW_X then dec(x) else dec(x,2);
  if y > 0 then if y < WINDOW_Y then dec(y) else dec(y,2);

  glUseProgram(SelShader.prog);
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glEnable(GL_SCISSOR_TEST);
  glScissor(x, y, 3, 3); // Die wenigsten Pixel mssen in den Buffer

  Render;
  // Kein SwapBuffers. Das codierte Bild sieht eh nicht hbsch aus.

  glUseProgram(0);
  glClearColor(0.1, 0.1, 0.1, 1.0);
  glDisable(GL_SCISSOR_TEST);

  glReadPixels(x, y, 3, 3, GL_RGBA, GL_UNSIGNED_BYTE, @PixelData[0]);

  count := 0;

  for i := 0 to 9 do
  begin
    if PixelData[4] = PixelData[i] then inc(count);

    if count > 3 then
    begin
      {$ifdef ENDIAN_BIG}
      result := (PixelData[4] shr 8)
      {$else}
      result := PixelData[4] and $FFFFFF;
      {$endif}
      exit;
    end;
  end;

  // Wenn diese Stelle erreicht wird, hat die Schleife nicht mehr als 3
  // gleichfarbige Pixel zusammenbekommen. Selektion ungltig.
  result := 0;
end;


//-----------------------------------------------------------------
// Fhrt ein paar Aufgaben aus, bevor es losgehen kann
//-----------------------------------------------------------------
procedure Setup;
var
  i, r: integer; // wird fr den Stern gebraucht
begin
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  // Vorsichtig mit Blend! Wenn Farben gemischt werden, oder der Alphakanal
  // zur Kodierung von Objekten genutzt wird, unbedingt in Select ausmachen!

  glEnable(GL_CULL_FACE);

  glClearColor(0.1, 0.1, 0.1, 1.0);

  glMatrixMode(GL_PROJECTION);
    glLoadIdentity;
    glOrtho(-400, 400, -300, 300, 0, 1);
  glMatrixMode(GL_MODELVIEW);

  // sel_vs.glsl und sel_fs.glsl laden
  SelShader.prog := SimpleProgram('data/sel');
  SelShader.color := glGetUniformLocation(SelShader.prog, 'color');
  SelShader.use_alpha_tex := glGetUniformLocation(SelShader.prog, 'use_alpha_tex');

  // Testtextur mit Alphakanal Laden und Binden
  LoadTGA('data/alphatex.tga');

  // Die Objekte mit Leben fllen - ausgelagert wegen bersicht
  with Objekte[1] do
  begin
    Name := 'Die Alphatextur';

    Displayliste := glGenLists(1);

    glNewList(Displayliste, GL_COMPILE);
      glBegin(GL_QUADS);
        glColor3f(1,1,1);
        glTexCoord2f(0,0);
        glVertex2f(-100, -100);
        glTexCoord2f(1,0);
        glVertex2f( 100, -100);
        glTexCoord2f(1,1);
        glVertex2f( 100,  100);
        glTexCoord2f(0,1);
        glVertex2f(-100,  100);
      glEnd;
    glEndList;

    posx := -250;
    posy := 0;
  end;

  with Objekte[2] do
  begin
    Name := 'Das bunte Dreieck';

    Displayliste := glGenLists(1);

    glNewList(Displayliste, GL_COMPILE);
      glBegin(GL_TRIANGLES);
        glColor3f(1,0,0);
        glVertex2f(-100, -100);
        glColor3f(0,1,0);
        glVertex2f( 100, -100);
        glColor3f(0,0,1);
        glVertex2f(  0,   100);
      glEnd;
    glEndList;

    posx := 250;
    posy := 0;
  end;

  with Objekte[3] do
  begin
    Name := 'Der rote Stern';

    Displayliste := glGenLists(1);

    glNewList(Displayliste, GL_COMPILE);
      glBegin(GL_TRIANGLE_FAN);
        glColor3f(0.7,0,0);
        glVertex3f(0,0,0);
        r := 100;

        for i := 0 to 16 do
        begin
          glVertex2f(cos(pi/8*i)*r, sin(pi/8*i)*r);
          if r = 100 then r := 60 else r := 100;
        end;
      glEnd;
    glEndList;

    posx := 25;
    posy := 70;
  end;
end;


//-----------------------------------------------------------------
// In der Eventschleife wird die Selektion ausgewertet
//-----------------------------------------------------------------
procedure MainLoop;
var
  event: TSDL_Event;
  done : boolean = false;
  i: longword;
begin
  repeat
    Render;
    SDL_GL_SwapBuffers;
    SDL_Delay(5);

    while SDL_PollEvent(@event)<>0 do
    begin
      case event.type_ of
        SDL_KEYDOWN: done := event.key.keysym.sym = SDLK_ESCAPE;
        SDL_QUITEV : done := true;

        SDL_MOUSEBUTTONDOWN:
        begin
          i := Select(event.button.x, event.button.y);

          if i = 0 then SDL_WM_SetCaption('Der Hintergrund', nil)
            else SDL_WM_SetCaption(pchar(Objekte[i].name), nil);
        end;
      end;
    end;
  until done;
end;


BEGIN
  SDL_Init(SDL_INIT_VIDEO);
  SDL_GL_SetAttribute(SDL_GL_MULTISAMPLESAMPLES, 16);
  // Antialiasing kann der fehlerkorrigierenden Selektion nichts anhaben

  SDL_SetVideoMode(WINDOW_X, WINDOW_Y, 0, SDL_OPENGL);

  InitOpenGL;
  ReadExtensions;

  Setup;
  MainLoop;

  SDL_Quit;
END.